/*
 * File: FileManager.h
 * Author: Adrian Łubik (2008)
 * License: GNU LGPL v3
 */

#ifndef FILEMANAGER_H
#define FILEMANAGER_H

#include <QString>
#include <QStringList>
#include <QDebug>
#include <QFileInfo>
#include <QFile>
#include <QDir>

/**
  * A class for finding and indexing files.
  */
class FileManager: public QObject
{
    Q_OBJECT
public:
    /**
      * Class constructor.
      */
    FileManager();

    /**
      * Class destructor.
      */
    virtual ~FileManager();


    /**
      * Searches for files (recurrently).
      * @param rootDir Searching root directory.
      * @param fileNamePattern File name pattern.
      * @param foundFiles List where found files' paths are stored.
      */   
    void findFiles( const QString rootDir, const QRegExp &fileNamePattern, QStringList &foundFiles );

    /**
      * Indexes files (only files) in a drirectory (recurrently).
      * @param rootDir Indexing root directory.
      */
    bool indexFiles( const QString &rootDir );

    /**
      * Searches for a file.
      * If a file has been found, function returns immediately.
      * This means that only first found file is saved to the result string.
      * @param rootDir Searching root directory.
      * @param fileName Name of the file to find.
      * @param foundFile String for path of found file.
      */ 
    static void findFile( const QString rootDir, const QString &fileName, QString &foundFile );

    /**
      * Copies files to a path.
      * @param filesToCopy List of files to copy to the path.
      * @param destPath Destination path of the files.
      */
    static void copyFiles( const QStringList &filesToCopy, const QString &destPath );

    /**
      * Copies a file to a path.
      * @param fileToCopy File to be copied.
      * @param destPath Destination path of the file.
      */
    static void copyFile( const QString &fileToCopy, const QString &destPath );

    /**
      * Removes a file.
      * @param fileName Name of file to be removed (full path).
      */
    static void removeFile( const QString &fileName );


    /**
      * Clears the index used by the indexFiles method.
      */
    void         clearIndex();

    /**
      * Gets the indexed files.
      * @return List of indexed files.
      */
    QStringList  getIndex();

    void clearFoundFilesList();

private:
    QStringList mAllFiles;
    QStringList mIndexedFiles;
};

#endif
