/*
 * File: Zypper.h
 * Author: Adrian Łubik (2008)
 * License: GNU LGPL v3
 */


#ifndef _ZYPPER_H_
#define _ZYPPER_H_

#include <QDebug>
#include <QString>
#include <QStringList>
#include <QByteArray>
#include <QRegExp>
#include <QProcess>
#include <QThread>

#include "MyProcess.h"
#include "Rpm.h"
#include "FileManager.h"

/**
  * @struct A struct representing the zypper repository data.
  */
struct Repository
{
    int number;
    bool enabled;
    bool refresh;
    QString type;
    QString alias;
    QString name;
};

typedef QList<Repository> Repositories;
typedef QListIterator<Repository> RepositoriesIterator;

/**
  * A class inherited from the abstract Rpm class.
  * The class is used to manage the zypper package manager.
  */
class Zypper: public Rpm
{
    Q_OBJECT
public:
    /**
      * Class constructor.
      * @param parent Parent's thread.
      */
    Zypper( QThread *parent = 0 );
    /**
      * Class destructor.
      */
    virtual ~Zypper();

    /**
      * Installs packages using the zypper process.
      * @param pkgs List of packages to install.
      * @return Exit code of the zypper process.
      */
    virtual int  installPackages( const QStringList &pkgs );

    /**
      * Adds a CD with packages to the list of repositories.
      * Checks whether the given volume is in fact a disc
      * with software. If it is, then tries to add it.
      * @param mntPoint The mounting point of the CD.
      * @return 0 if everything wend ok.
      */ 
    virtual int  addCD( const QString &mntPoint );

    /**
      * Checks whether the given packages are available to install.
      * Uses the apt-cache process to get informations about packages.
      * @param pkgs List of packages to check.
      * @return True if all of the given packages are available.
      */
    virtual bool getArePackagesAvailable( const QStringList &pkgs );

    /**
      * Adds a repository to the list of repositories on the system.
      * @param path Path to the repository (ex. a CD mount point).
      * @param alias Alias of the repository to recognise it by users.
      * @return Exit code of the zypper process.
      */
    int  addRepo( const QString &path, const QString &alias );

    /**
      * Enables or disables a repository.
      * @param alias Alias of the repository to be enabled/disabled.
      * @param enable True - enable repository, false - disable repository.
      * @return True if the operation has been performed successfully.
      */
    bool enableRepo( QString alias, bool enable );

    /**
      * Enable a repository.
      * @param alias Alias of the repository to be enabled.
      */
    bool enableRepo( const QString &alias ) {
        return enableRepo( alias, true );
    }

    /**
      * Disable a repository.
      * @param alias Alias of the repository to be disabled.
      */
    bool disableRepo( const QString &alias ) {
        return enableRepo( alias, false );
    }

    /**
      * Get system repositories.
      * @return List of the Repository struct of system repositories.
      */
    Repositories getRepos();

private slots:
    void debugProcess( QByteArray ba );
    void debugZyppRepos( QByteArray ba );

private:
    QByteArray dbg;
    MyProcess *m_installProcess;
    MyProcess *m_zypperProcess;
    QProcess  *m_addRepoProcess;
    QProcess  *m_modifyRepoProcess;
    QProcess  *m_getAvailableProcess;

    Repositories repos;
    QRegExp      repoRegExp;
    QRegExp      processing;
    QThread     *m_parentThread;
};

#endif

