/*
 * File: UbuDSL_Thread.cpp
 * Created: Thu 07 Feb 2008
 * Last modified: Tue 23 Apr 2008
 * Author: Adrian Łubik
 * License: GNU GPL v3 or newer
 */

#include "UbuDSL_Thread.h"
#include <unistd.h>

extern bool  debug;
extern FILE *pFile;

static const int CONFIG_VERSION = 100277;

UbuDSL_Thread::UbuDSL_Thread( const QString &login, const QString &password, const int &protocol,
                              const int &VPI, const int &VCI, const QString &service,
                              const QStringList &dnsServers,  const QPair<int, int> &virtualDeviceID,
                              const bool &connectAtBoot,      const bool &connectAtModemPlug,
                              const int lockedFD, EndState *endState
                            )
        : fileManager( new FileManager() ),
        scriptManager( new ScriptManager() ),
        modemManager( new ModemManager( virtualDeviceID ) ),
        mediaManager( new MediaManager( this ) ),
        m_lockedFD( lockedFD ),
        volumeMountPoint( QString() ),
        volumesCount( 0 ),
        daemonSettings( new QSettings( UBUDSL_PREFIX + "/share/ubudsl/.daemon_ng.ini", QSettings::IniFormat ) ),
        m_sys( UbuDSL::SystemInfo::getDistributionName() ),
        m_ver( UbuDSL::SystemInfo::getDistributionRelease() ),
        m_dnsServers( dnsServers ),
        m_canExit( true ),
        m_endState( endState )
{
    qDebug() << "UbuDSL_Thread::UbuDSL_Thread(): ctor called.";

    this->login         = login;
    this->password      = password;
    this->protocol      = protocol;
    this->VPI           = VPI;
    this->VCI           = VCI;

    qDebug() << "UbuDSL_Thread::UbuDSL_Thread(): parameters:\n"
    << "  * Service:" << service << "\n"
    << "  * Protocol:" << ( protocol == 0 ? "PPPoA\n" : "PPPoE\n" )
    << "  * VPI:" << VPI << "\n"
    << "  * VCI:" << VCI;

    daemonSettings->setValue( "protocol", ( protocol == 0 ? "PPPoA" : "PPPoE" ) );
    daemonSettings->setValue( "VPI", VPI );
    daemonSettings->setValue( "VCI", VCI );
    daemonSettings->setValue( "service", service );
    daemonSettings->setValue( "connect_at_boot", connectAtBoot );
    daemonSettings->setValue( "connect_at_modem_plug", connectAtModemPlug );
    daemonSettings->setValue( "version", CONFIG_VERSION );

    QString sysIni = UBUDSL_PREFIX + "/share/ubudsl/configs/system.ini";
    if ( !QFile::exists( sysIni ) )
    {
        qWarning() << sysIni << "file not found!";
        systemIni = new QSettings( ":configs/system.ini", QSettings::IniFormat );
    }
    else
    {
        systemIni = new QSettings( sysIni, QSettings::IniFormat );
    }

    if ( endState != NULL )
        *endState = Success;

    systemIni->beginGroup( m_sys + "_" + m_ver );

//    packageManager = new PackageManager ( systemIni->value ( "package_manager", "" ).toString(), this );
    QString tmp( systemIni->value( "package_manager", "" ).toString() );

    if ( tmp == "apt" )
    {
        packageManager = new Apt( this );
    }
    else if ( tmp == "zypper" )
    {
        packageManager = new Zypper( this );
    }
    else
    {
        qFatal( tmp.isEmpty() ? "No package manager specified, aborting!"
                : "Unknown package manager: %s", tmp.toUtf8().constData() );
    }

    packagesList.clear();
    packagesToInstall.clear();
    cdIndex.clear();

    ModemManager::initDriversMap( m_drivers );
}


UbuDSL_Thread::~UbuDSL_Thread()
{
    qDebug() << "UbuDSL_Thread::~UbuDSL_Thread(): dtor called.";

    delete fileManager;
    delete packageManager;
    delete mediaManager;
    delete scriptManager;

    qDebug() << "UbuDSL_Thread::~UbuDSL_Thread(): dtor finished.";
}


void UbuDSL_Thread::run()
{
    emit message( tr( "Getting system informations..." ) );

    QString kernel   = UbuDSL::SystemInfo::getKernelRelease();
    QString arch     = UbuDSL::SystemInfo::getArchitecture();

    qDebug() << "UbuDSL_Thread::run(): system information:\n"
    << "  * Detected system:" << m_sys << m_ver << "\n"
    << "  * Detected architecture:" << arch << "\n"
    << "  * Detected kernel:" << kernel;

    /*QRegExp kernelRegExp ( "([0-9]{1})\\.([0-9]{1})\\.([0-9]{1,2})" );
    int kernelNum = 0;

    if ( kernel.contains ( kernelRegExp ) )
    {
        QString str = kernelRegExp.cap ( 1 );
        str += kernelRegExp.cap ( 2 );
        str += kernelRegExp.cap ( 3 );

        kernelNum = str.toInt();
    }*/

    emit message( QString( tr( "  Kernel: %1" ) ).arg( kernel ) );
    emit message( QString( tr( "  Architecture: %1" ) ).arg( arch ) );
    emit message( QString( tr( "  System: %1 %2" ) ).arg( m_sys ).arg( m_ver ) );

    emit detectingModem();
    emit message( tr( "Please plug your modem in" ) );

    modemManager->connectToDBus();

    connect( modemManager, SIGNAL( foundModem() ), this, SLOT( quit() ) );    //after modem is found, quit from waiting loop
    exec();
    disconnect( modemManager, SIGNAL( foundModem() ), this, SLOT( quit() ) );

    emit modemDetected( modemManager->getModemName() );

    qDebug() << "UbuDSL_Thread::run(): detected modem:" << modemManager->getModemName();

    daemonSettings->setValue( "modem",      modemManager->getModemName() );
    daemonSettings->setValue( "vendor_id",  modemManager->getVendorID() );
    daemonSettings->setValue( "product_id", modemManager->getProductID() );
    daemonSettings->setValue( "udi",        modemManager->getUDI() );
    driverName     = modemManager->getDriverName();
    driverVersion  = modemManager->getDriverVersion();
    m_driver = driverName + "-" + driverVersion;

    delete modemManager; //because it is no more needed

    if ( m_lockedFD > 0 )
        unlockFile( m_lockedFD );

    qDebug() << "UbuDSL_Thread::run(): checking basic packages...";

    emit message( tr( "Checking basic packages..." ) );
    QStringList notInstalledBasic;

    if ( !systemIni->value( "basic_from_cd", QString() ).toString().isEmpty() )
        notInstalledBasic = packageManager->getNotInstalled( systemIni->value( "basic_from_cd", QStringList() ).toStringList() );

    if ( notInstalledBasic.isEmpty() )
        qDebug() << "UbuDSL_Thread::run(): not installed basic packages list is empty...";
    else
        qDebug() << "UbuDSL_Thread::run(): not installed basic packages list:" << notInstalledBasic;

    if ( !notInstalledBasic.isEmpty() && volumeMountPoint.isEmpty() )
    {
        qDebug() << "UbuDSL_Thread::run(): waiting for disc (001).";

        waitForDisc();
    }

    if ( !packageManager->getArePackagesAvailable( notInstalledBasic ) )
    {
        packageManager->addCD( volumeMountPoint );
    }

    m_canExit = false;

    if ( !notInstalledBasic.isEmpty() )
    {
        qDebug() << "UbuDSL_Thread::run(): installing basic packages...";
        emit message( tr( "Installing basic packages..." ) );

        if ( packageManager->installPackages( notInstalledBasic ) == 0 )
        {
            qDebug() << "UbuDSL_Thread::run(): basic packages from CD installed successfully.";
            emit message( tr( "Basic packages from CD installed successfully" ) );
        }
        else
        {
            qDebug() << "UbuDSL_Thread::run(): basic packages from CD installation failed!";
            emit message( tr( "Basic packages from CD installation failed!" ) );
        }
    }


    if ( m_driver == "unicorn-neo" && QProcess::execute( "modprobe unicorn_usb_eth" ) != 0 )
    {
        qDebug() << "UbuDSL_Thread::run(): checking development packages...";

        emit message( tr( "Checking development packages..." ) );
        QStringList notInstalledDevel
        ;
        if ( !systemIni->value( "devel_from_cd", QString() ).toString().isEmpty() )
            notInstalledDevel = packageManager->getNotInstalled( systemIni->value( "devel_from_cd", QStringList() ).toStringList() );


        if ( notInstalledDevel.isEmpty() )
            qDebug() << "UbuDSL_Thread::run(): not installed development packages list is empty...";
        else
            qDebug() << "UbuDSL_Thread::run(): not installed development packages list:" << notInstalledDevel;

        if ( !notInstalledDevel.isEmpty() && volumeMountPoint.isEmpty() )
        {
            qDebug() << "UbuDSL_Thread::run(): waiting for disc (002).";

            waitForDisc();
        }

        if ( !packageManager->getArePackagesAvailable( notInstalledDevel ) )
        {
            packageManager->addCD( volumeMountPoint );
        }

        if ( !notInstalledDevel.isEmpty() )
        {
            qDebug() << "UbuDSL_Thread::run(): development basic packages...";
            emit message( tr( "Installing development packages..." ) );

            if ( packageManager->installPackages( notInstalledDevel ) == 0 )
            {
                qDebug() << "UbuDSL_Thread::run(): development packages from CD installed successfully.";
                emit message( tr( "Development packages from CD installed successfully" ) );
            }
            else
            {
                qDebug() << "UbuDSL_Thread::run(): development packages from CD installation failed!";
                emit message( tr( "Development packages from CD installation failed!" ) );
            }
        }

        qDebug() << "UbuDSL_Thread::run(): preparing the unicorn driver...";
        emit message( tr( "Preparing driver..." ) );

        ModemManager::ModemDriver driver;
        QListIterator<ModemManager::ModemDriver> it( m_drivers );

        while ( it.hasNext() )
        {
            driver = it.next();

            if ( driver.name == driverName && driver.version == driverVersion )
            {
                qDebug() << "\n\n\n" << arch << driver.amd64_compatible;
                if ( arch == "amd64" && !driver.amd64_compatible )
                {
                    qDebug() << "UbuDSL_Thread::run(): incompatible architecture!";
                    //emit incompatibleArch();

                    if ( m_endState != NULL )
                        *m_endState = IncompatibleArchitecture;

                    return;
                }
                if ( installPrecompiledDriver( driver ) )
                    break;

                Compiler *compiler = new Compiler( driver, 0 );
                qDebug() << "UbuDSL_Thread::run(): backuping driver.";
                compiler->backupDriver();
                qDebug() << "UbuDSL_Thread::run(): extracting driver.";
                compiler->extractDriver();
                qDebug() << "UbuDSL_Thread::run(): patching driver (if needed).";
                compiler->patchDriver();

                emit message( tr( "  Compiling driver..." ) );
                qDebug() << "UbuDSL_Thread::run(): compiling driver.";

                if ( compiler->compileDriver() == 0 )
                {
                    emit message( tr( "  Installing driver" ) );
                    qDebug() << "UbuDSL_Thread::run(): installing driver.";

                    if ( compiler->installDriver() != 0 )
                    {
                        emit message( tr( "    Driver installation failed!" ) );
                        emit message( tr( "    Restoring old driver..." ) );

                        qDebug() << "UbuDSL_Thread::run(): installation failed, restoring driver.";

                        compiler->restoreDriver();

                        if ( m_endState != NULL )
                            *m_endState = Failed;
                    }
                }
                else
                {
                    compiler->restoreDriver();
                    emit message( tr( "    Driver compilation failed!" ) );

                    qDebug() << "UbuDSL_Thread::run(): installation failed, restoring driver.";

                    if ( m_endState != NULL )
                        *m_endState = Failed;
                }
                delete compiler;
                break;
            }
        }
    }

    qDebug() << "UbuDSL_Thread::run(): installing firmware.";

    emit message( tr( "Installing firmware..." ) );
    Firmware *fw = new Firmware;
    if ( !fw->installUeagleFirmware( UBUDSL_PREFIX + "/share/ubudsl/firmware/ueagle-firmware.tar" ) )
        emit message( tr( "    Failed to install firmware: ueagle-atm!" ) );
    if ( !fw->installSpeedtchFirmware( UBUDSL_PREFIX + "/share/ubudsl/firmware/speedtch-firmware.tar" ) )
        emit message( tr( "    Failed to install firmware: speedtch!" ) );
    if ( !fw->installCxacruFirmware( UBUDSL_PREFIX + "/share/ubudsl/firmware/cxacru-firmware.tar" ) )
        emit message( tr( "    Failed to install firmware: cxacru!" ) );
    delete fw;

    m_canExit = true;

    qDebug() << "UbuDSL_Thread::run(): creating scripts.";
    emit message( tr( "Creating scripts..." ) );

    if ( !scriptManager->writeOptionsScript() )
    {
        qDebug() << "UbuDSL_Thread::run(): could not write ppp options script!";
        emit message( tr( "  Could not write ppp options script!" ) );
    }
    if ( !scriptManager->writeSecretScripts( login, password ) )
    {
        qDebug() << "UbuDSL_Thread::run(): could not write authorization scripts!";
        emit message( tr( "  Could not write authorization scripts!" ) );
    }
//    if ( !scriptManager->writeUnicornScript ( login ) )
//    {
//        qDebug() << "UbuDSL_Thread::run(): could not write unicorn connection script!";
//        emit message ( tr ( "  Could not write unicorn connection script!" ) );
//    }

    QString modprobeOptions( systemIni->value( "modprobe_path", "/etc/modprobe.d/options" ).toString() );

    if ( protocol == 0 )
    {
        scriptManager->writeUnicornOption( modprobeOptions, VPI, VCI, protocol );
    }
    else
    {
        scriptManager->writeUnicornOption( modprobeOptions, VPI, VCI, protocol );
    }

    if ( !scriptManager->writePPPoAScript( VPI, VCI, login ) )
    {
        qDebug() << "UbuDSL_Thread::run(): could not write PPPoA connection script!";
        emit message( tr( "  Could not write PPPoA connection script!" ) );
    }
    if ( !scriptManager->writePPPoEScript( login ) )
    {
        qDebug() << "UbuDSL_Thread::run(): could not write PPPoE connection script!";
        emit message( tr( "  Could not write PPPoE connection script!" ) );
    }

    scriptManager->setScriptsPermissions();

    DNS_Manager dns;

    qDebug() << "DNS:" << m_dnsServers;

    if ( !dns.setServers( m_dnsServers ) /*|| !dns.isFileCorrect()*/ )
    {
        qDebug() << "UbuDSL_Thread::run(): setting OpenDSN as the default DNS servers";
        emit message( tr( "Setting OpenDSN as the default DNS servers..." ) );

        QStringList OpenDNS( QStringList() << "208.67.222.222" << "208.67.220.220" );

        dns.setServers( OpenDNS );
        daemonSettings->setValue( "DNS", OpenDNS );
    }
    else
    {
        qDebug() << "UbuDSL_Thread::run(): setting custom DNS servers:" << m_dnsServers;
        daemonSettings->setValue( "DNS", m_dnsServers );
    }


    emit message( tr( "Loading the driver..." ) );

    if ( driverName == "unicorn" )
        driverName = "unicorn_usb_eth";

    QProcess::execute( "modprobe", QStringList() << "-r" << "-f" << driverName );
    QProcess::execute( "modprobe", QStringList() << driverName );

    emit message( tr( "Restarting the UbuDSL Daemon..." ) );

    QProcess::execute( "/etc/init.d/ubudsld restart" );

    quit();
}


void UbuDSL_Thread::volumesListChanged()
{
    int count = mediaManager->getVolumesList().count();
    if ( count != volumesCount )
    {
        volumesCount = count;

        volumes = mediaManager->getVolumesList();
        emit setVolumesList( volumes );
    }
}


void UbuDSL_Thread::volumeSelected( const int &index )
{
    if ( index == -1 ) //ex. called from gui
    {
        qDebug() << "UbuDSL_Thread::volumeSelected(): index == -1, skipping.";
        emit discMounted();
    }

    if ( index > volumes.count() - 1 || index < 0 )
    {
        qDebug() << "UbuDSL_Thread::volumeSelected(): index > volumes.count() - 1 || index < 0. (volumes.count:" << volumes.count() <<  ").";
        return;
    }

    if ( mediaManager->getVolumeIsMounted( volumes[index].getUdi() ) )
    {
        qDebug() << "UbuDSL_Thread::volumeSelected(): volume is mounted.";
        volumeMountPoint = mediaManager->getVolumeMountPoint( volumes[index].getUdi() );
    }
    else
    {
        qDebug() << "UbuDSL_Thread::volumeSelected(): volume is NOT mounted.";
        QString mntPoint = "/tmp/ubudsl_mnt";
        int mountingResult = mediaManager->mountVolume( volumes[index].getUdi(), mntPoint );

        if ( mountingResult == 0 )
        {
            qDebug() << "UbuDSL_Thread::volumeSelected(): volume mounted in" << mntPoint;
            volumeMountPoint = mntPoint;
        }
        else
        {
            qDebug() << "UbuDSL_Thread::volumeSelected(): could NOT mount volume in" << mntPoint;
            volumeMountPoint = QString();
        }
    }

    emit discMounted();
}


void UbuDSL_Thread::waitForDisc()
{
    emit message( tr( "Waiting for disc..." ) );
    emit setVolumesList( mediaManager->getVolumesList() );
    emit waitingForDisc();
    volumes = mediaManager->getVolumesList();
    connect( mediaManager, SIGNAL( volumeAdded() ), this, SLOT( volumesListChanged() ) );
    connect( mediaManager, SIGNAL( volumeRemoved() ), this, SLOT( volumesListChanged() ) );
    connect( this, SIGNAL( discMounted() ), this, SLOT( quit() ) );
    exec();
    disconnect( this, SIGNAL( discMounted() ), this, SLOT( quit() ) );
    disconnect( mediaManager, SIGNAL( volumeAdded() ), this, SIGNAL( volumesListChanged() ) );
    disconnect( mediaManager, SIGNAL( volumeRemoved() ), this, SIGNAL( volumesListChanged() ) );
}


bool UbuDSL_Thread::getCanExit()
{
    return m_canExit;
}

bool UbuDSL_Thread::installPrecompiledDriver( const ModemManager::ModemDriver &driver )
{
    QString driverFile = UBUDSL_PREFIX +
                                 "/lib/ubudsl/drivers/precompiled/" +
                                 UbuDSL::SystemInfo::getDistributionName() + "_" +
                                 UbuDSL::SystemInfo::getDistributionRelease() + "_" +
                                 UbuDSL::SystemInfo::getArchitecture() + "_" +
                                 UbuDSL::SystemInfo::getKernelRelease() + "_" +
                                 driver.compiled_driver_name;

    if ( QFile::exists( driverFile ) )
    {
        //emit message( tr( "  Installing precompiled driver..." ) );
        qDebug() << "UbuDSL_Thread::installPrecompiledDriver(): Installing precompiled driver.";

        QDir d( "/lib/modules/" + UbuDSL::SystemInfo::getKernelRelease() + "/extra" );
        if ( !d.exists() )
        {
            d.mkdir( d.absolutePath() );
        }

        qDebug() << "d exists:" << d.exists();
        qDebug() << "d path:" << d.absolutePath();
        qDebug() << "d name" << driver.compiled_driver_name;

        QString destinationPath( d.absolutePath() + "/" + driver.compiled_driver_name );

        if ( QFile::copy( driverFile, destinationPath ) )
        {
            if ( chown( QFile::encodeName( destinationPath ), 0, 0 ) != 0 )
                qDebug() << "UbuDSL_Thread::installPrecompiledDriver(): failed to change module owner!";

            QFile( destinationPath ).setPermissions( QFile::ReadOwner  |
                                                     QFile::WriteOwner |
                                                     QFile::ReadUser   |
                                                     QFile::ReadGroup  |
                                                     QFile::ReadOther  );

            QProcess::execute( "depmod -aq" );
            //emit message( tr( "  Precompiled driver installed successfully..." ) );
            qDebug() << "UbuDSL_Thread::installPrecompiledDriver(): Precompiled driver installed successfully.";
            return true;
        }
    }
    else
    {
        qDebug() << "UbuDSL_Thread::installPrecompiledDriver(): No precompiled driver found!";
    }

    return false;
}
