#include "CompilerThread.h"
#include "Globals.h"

CompilerThread::CompilerThread( const UbuDSL_Daemon *, const QString &driverName, QObject *parent )
        : QThread( parent ),
        m_driverName( driverName ),
        m_result( Success )
{
    writeLog( "CompilerThread::CompilerThread(): called." );
    ModemManager::initDriversMap( m_drivers );
    writeLog( "CompilerThread::CompilerThread(): finished." );
}

CompilerThread::~CompilerThread()
{
    writeLog( "CompilerThread::~CompilerThread(): called and finished." );
}

void CompilerThread::setDriverName( const QString &driverName )
{
    m_driverName = driverName;
}

int CompilerThread::testDriver()
{
    writeLog( "CompilerThread::testDriver(): called." );

    m_result = DriverTested;

    if ( m_driverName.isEmpty() )
    {
        qDebug() << "CompilerThread::testDriver(): m_driverName is empty!";
        return -1;
    }

    int ret = QProcess::execute( "modprobe", QStringList() << m_driverName );
    writeLog( "CompilerThread::testDriver(): finished." );
    return ret;
}

void CompilerThread::run()
{
    writeLog( "CompilerThread::run(): called." );

    if ( m_driverName.isEmpty() )
    {
        qDebug() << "CompilerThread::run(): m_driverName is empty!";
        return;
    }

    ModemManager::ModemDriver driver;
    QListIterator<ModemManager::ModemDriver> it( m_drivers );

    while ( it.hasNext() )
    {
        driver = it.next();

        if ( driver.compiled_driver_name == ( m_driverName + ".ko" ) )
        {
            if ( installPrecompiledDriver( driver ) )
            {
                writeLog( "CompilerThread::run(): a precompiled driver has been successfully installed." );
                break;
            }

            Compiler *compiler = new Compiler( driver, 0 );
            compiler->backupDriver();
            compiler->extractDriver();
            compiler->patchDriver();

            writeLog( "CompilerThread::run(): compiling driver." );

            if ( compiler->compileDriver() == 0 )
            {
                writeLog( "CompilerThread::run(): installing driver." );

                if ( compiler->installDriver() == 0 )
                {
                    m_result = Success;
                    testDriver();
                }
                else {
                    writeLog( "CompilerThread::run(): driver installation failed!" );
                    m_result = InstallationFailed;
                    compiler->restoreDriver();
                }
            }
            else
            {
                compiler->restoreDriver();
                writeLog( "CompilerThread::run(): driver compilation failed!" );
                m_result = CompilationFailed;
            }

            delete compiler;
            break;
        }
    }

    writeLog( "CompilerThread::run(): finished." );
}

int CompilerThread::getResult()
{
    return m_result;
}

bool CompilerThread::installPrecompiledDriver( const ModemManager::ModemDriver &driver )
{
    QString driverFile = UBUDSL_PREFIX +
                                 "/lib/ubudsl/drivers/precompiled/" +
                                 UbuDSL::SystemInfo::getDistributionName() + "_" +
                                 UbuDSL::SystemInfo::getDistributionRelease() + "_" +
                                 UbuDSL::SystemInfo::getArchitecture() + "_" +
                                 UbuDSL::SystemInfo::getKernelRelease() + "_" +
                                 driver.compiled_driver_name;

    if ( QFile::exists( driverFile ) )
    {
        writeLog( "CompilerThread::installPrecompiledDriver(): Installing precompiled driver." );

        QDir d( "/lib/modules/" + UbuDSL::SystemInfo::getKernelRelease() + "/extra" );
        if ( !d.exists() )
        {
            d.mkdir( d.absolutePath() );
        }

        QString destinationPath( d.absolutePath() + "/" + driver.compiled_driver_name );

        if ( QFile::copy( driverFile, destinationPath ) )
        {
            if ( chown( QFile::encodeName( destinationPath ), 0, 0 ) != 0 )
                writeLog( "CompilerThread::installPrecompiledDriver(): failed to change module owner!" );

            QFile( destinationPath ).setPermissions( QFile::ReadOwner  |
                                                     QFile::WriteOwner |
                                                     QFile::ReadUser   |
                                                     QFile::ReadGroup  |
                                                     QFile::ReadOther  );

            QProcess::execute( "depmod -aq" );
            writeLog( "CompilerThread::installPrecompiledDriver(): Precompiled driver installed successfully." );
            return true;
        }
        else
            writeLog( "CompilerThread::installPrecompiledDriver(): Failed to copy the driver..." );
    }
    else
    {
        writeLog( "CompilerThread::installPrecompiledDriver(): No precompiled driver found!" );
    }

    return false;
}
